# -*- coding: utf-8 -*-
"""

Regelstrecke mit einer wechselnden Zeitkonstanten und PI-Regler, die 
Regler-Parameter werden in einem Wertebereich variert (Batch-Betrieb) 
und die Güte in einem 3D-Plot dargestellt. 
Die optimalen Parameter sind im Minimum zu finden

Created on 2.12.  2021 
@author: philippsen
""" 

import numpy as np
import control
import control.matlab
import matplotlib.pyplot as plt
plt.rcParams['pdf.fonttype'] = 42
plt.rcParams['ps.fonttype'] = 42
from matplotlib import cm
from mpl_toolkits.mplot3d import Axes3D

n = 400 # Anzahl Schritte
T0 = 0.02 # Abtastzeit
kk = 30  # Anzahl Parametersätze
xx = np.zeros(n)    # Regelgrößenvektor
u = np.zeros(n)     # Stellgrößenvektor
tk= np.arange(0,T0*n + T0,T0)    # k-Achse
MKp = np.zeros((kk,kk))
dKp_kk = 3/kk; dTn_kk = 3/kk # Änderungsstufen Parameter
MTn = np.zeros((kk,kk))
Q1 = np.zeros((kk,kk)); Q2 = np.zeros((kk,kk)) # Güte-Matrizen
Qs = np.zeros((kk,kk)) # Summe Güte-Matrizen
#Erzeugung untere und ober Grenzlinie
nn = len(tk); glT1 = 2.4; glT2 = 3.1; tolband = 0.03
gl = np.zeros(nn) # Grenzlinie untere
# i Werte für glT1 und glT2
w = 1.0    #Sollwert
glT1i = int(glT1/T0); glT2i = int(glT2/T0)
sh = (w - tolband)/(glT2i - glT1i) #Stufenhöhe

for i in range(glT1i, nn): # Einträge Werte in gl
    gl[i] = (i - glT1i)*sh
    if i > glT2i:
        gl[i] = w - tolband


pt1 = control.tf(1.0, [0.2, 1.0])
strecke1 = control.tf([1],[1, 1]) *pt1  # Strecke 1
strecke2 = control.tf([1],[2, 1]) *pt1  # Strecke 2

for ii in range(0, kk): # Liste von 0 bis kk-1
    for jj in range(0, kk):
        Kp = 1.0 + dKp_kk*(ii + 1); Tn = 0.0 + dTn_kk*(jj +1)      # Reglerparameter
        reg = control.tf([Kp*Tn, Kp], [Tn, 0])
        MKp[ii,jj] = Kp; MTn[ii,jj] = Tn
        G0 = reg*strecke1
        Gw = control.feedback(G0, 1) # % Regelkreis
        xk , tt = control.matlab.step(Gw,tk)
        drin = True
        for i in range(glT1i, nn): # Überschwingen nicht zulassen ab 0, sonst ab T2
            if xk[i] > (w+tolband) or xk[i] < gl[i]:
                drin = False
                g = 1000
                # Die untere Grenzlinie könnte durch w-tolband ersetzt werden!
        
        if drin:
            Gy = control.feedback(reg, strecke1) # % Stellgröße
            yk , tt = control.matlab.step(Gy,tk)
            g = np.dot(yk,yk)
        Q1[ii,jj] = g # Stellenergie
        
        G0 = reg*strecke2
        Gw = control.feedback(G0, 1) # % Regelkreis
        xk , tt = control.matlab.step(Gw,tk)
        drin = True
        for i in range(glT1i, nn): # Überschwingen nicht zulassen ab 0, sonst ab T2
            if xk[i] > (w+tolband) or xk[i] < gl[i]:
                drin = False
                g = 1000
                # Die untere Grenzlinie könnte durch w-tolband ersetzt werden!
        
        if drin:
            Gy = control.feedback(reg, strecke2) # % Stellgröße
            yk , tt = control.matlab.step(Gy,tk)
            g = np.dot(yk,yk)*T0
        Q2[ii,jj] = g # Stellenergie


print('Minimum1 = ',Q1.min(),'Anzahl Iterationen = ',kk*kk )
pa = np.where(Q1 == Q1.min())
print('Kp = ',MKp[pa[0],pa[1]],'   Tn = ',MTn[pa[0],pa[1]] )
print('Minimum2 = ',Q2.min(),'Anzahl Iterationen = ',kk*kk )
pa = np.where(Q2 == Q2.min())
print('Kp = ',MKp[pa[0],pa[1]],'   Tn = ',MTn[pa[0],pa[1]] )
Qs = Q1 + Q2
print('MinimumSumme = ',Qs.min(),'Anzahl Iterationen = ',kk*kk )
pa = np.where(Qs == Qs.min())
print('Kp = ',MKp[pa[0],pa[1]],'   Tn = ',MTn[pa[0],pa[1]] )
fig = plt.figure()
ax = fig.add_subplot(projection='3d')
ax.plot_surface(MKp, MTn, Q1, rstride=1, cstride=1, cmap=cm.viridis)
plt.xlabel('Kp'); plt.ylabel('Tn'); ax.set_zlabel('Q1')
fig = plt.figure()
ax = fig.add_subplot(projection='3d')
ax.plot_surface(MKp, MTn, Q2, rstride=1, cstride=1, cmap=cm.viridis)
plt.xlabel('Kp'); plt.ylabel('Tn'); ax.set_zlabel('Q2')

K = MKp[pa[0],pa[1]][0]; Tn = MTn[pa[0],pa[1]][0]
R = control.matlab.tf([K*Tn, K],[Tn, 0.0])    
G0 = R*strecke1
Gw = control.matlab.feedback(G0,1)
x, t = control.matlab.step(Gw,tk)
stell = control.matlab.feedback(R, strecke1)
y, t = control.matlab.step(stell,tk)

G0 = R*strecke2
Gw = control.matlab.feedback(G0,1)
x2, t = control.matlab.step(Gw,tk)
stell2 = control.matlab.feedback(R, strecke2)
y2, t = control.matlab.step(stell2,tk)

fig, ax1 = plt.subplots()
ax1.plot(t, x,'b', t,x2,"b--") 
simEnde = t.max()
ax1.plot([0, simEnde], [w+tolband, w+tolband], 'g--') 
ax1.plot([glT2, simEnde], [w-tolband, w-tolband], 'g--') 

ax1.set_ylabel('x x2--', color="blue", fontsize=14) 
plt.grid()
ax2 = ax1.twinx() 
ax2.plot(t, y,'r', t,y2,"r--") 
ax2.set_ylabel('y y2--', color="red", fontsize=14) 
ax1.set_xlabel('t [s]', fontsize=12) 
s = str("Robuster Entwurf vollst. Parameterraum Kr = %1.2f "% K)
s2 = str("Tn = %1.2f" % Tn)
plt.title(s+s2)

